"""Notes about race condition:
We use soft locks, which watches .lock file to check if appears and disappears.

AbsPath:
    Almost race cond free.
    Fast polling (0.01sec)

GCSURI:
    Default slow polling (10sec)
    Python API is not thread-safe.
    API provides a way to lock an object.

S3URI:
    Default slow polling (10sec)

HTTPURL:
    Read-only storage so don't need to test

Test nth threads competing to write on the same file v6.txt.
Compare written string vs read string.

Important notes:
    Python API for GCS client() is not thread-safe.
        So we need to specify thread_id here.
        URIBase (and its child GCSURI) has a thread_id
        This will make a new GCS client instance for each thread.
    S3 Object Lock is based on Versioning.
        We don't allow versioning so keep using unstable soft file lock.
"""
import os
from multiprocessing import Pool

from autouri.autouri import AutoURI

from .files import v6_txt_contents


def write_v6_txt(x):
    """Lock -> write_lockfree -> read -> compare written vs read -> unlock.
    This writes different text for different thread.
    """
    uri, i = x
    s = v6_txt_contents() + str(i)
    u = AutoURI(uri, thread_id=i)

    with u.get_lock(no_lock=False):
        u.write(s, no_lock=True)
        assert u.read() == s


def run_write_v6_txt(prefix, nth):
    s = os.path.join(prefix, "v6.txt")
    u = AutoURI(s)
    if u.exists:
        u.rm()
    p = Pool(nth)
    p.map(write_v6_txt, list(zip([s] * nth, range(nth))))
    p.close()
    p.join()


def test_race_cond_autouri_write_local(local_test_path):
    prefix = os.path.join(local_test_path, "test_race_cond_autouri_write_local")
    nth = 10
    run_write_v6_txt(prefix, nth)


def test_race_cond_autouri_write_gcs(gcs_test_path):
    prefix = os.path.join(gcs_test_path, "test_race_cond_autouri_write_gcs")
    nth = 5
    run_write_v6_txt(prefix, nth)


def test_race_cond_autouri_write_s3(s3_test_path):
    nth = 3
    prefix = os.path.join(s3_test_path, "test_race_cond_autouri_write_s3")
    run_write_v6_txt(prefix, nth)



"""Make sure to run these tests on a machine with correct os time
"""
from datetime import datetime, timedelta, timezone
from unittest.mock import patch

from autouri.ntp_now import get_cached_offset, now_utc, reset_cached_offset


def test_now_utc():
    reset_cached_offset()

    assert abs((now_utc() - datetime.now(timezone.utc)).total_seconds()) < 0.05


class MockedDataTime(datetime):
    @classmethod
    def now(cls, timezone):
        return datetime.now(timezone) - timedelta(0, 25)


def test_now_utc_wrong_os_time():
    reset_cached_offset()
    with patch("autouri.ntp_now.datetime", MockedDataTime):
        ntp_now_utc = now_utc()

    # should be accurate even though system time is 25 second behind NTP server time
    assert abs((ntp_now_utc - datetime.now(timezone.utc)).total_seconds()) < 0.05

    # cache offset should be 25 second
    cached_offset_in_seconds = get_cached_offset().total_seconds()
    assert cached_offset_in_seconds > 24.95 and cached_offset_in_seconds < 25.05



"""Race condition test will be done in test_race_cond.py
"""
import time

from filelock import BaseFileLock, Timeout

from autouri.autouri import AutoURI, URIBase


class AutoURIFileLockTestException(Exception):
    pass


def test_autouri_lock(local_v6_txt, gcs_v6_txt, s3_v6_txt):
    for v6_txt in (local_v6_txt, gcs_v6_txt, s3_v6_txt):
        u_lock = AutoURI(v6_txt + URIBase.LOCK_FILE_EXT)
        lock = AutoURI(v6_txt).get_lock(no_lock=False)

        lock.acquire()
        try:
            assert u_lock.exists
            time.sleep(1)
        finally:
            lock.release()
        assert not u_lock.exists

        # trivial dummy lock
        lock = AutoURI(v6_txt).get_lock(no_lock=True)
        assert not isinstance(lock, BaseFileLock)


def test_autouri_lock_raise(local_v6_txt, gcs_v6_txt, s3_v6_txt):
    for v6_txt in (local_v6_txt, gcs_v6_txt, s3_v6_txt):
        u_lock = AutoURI(v6_txt + URIBase.LOCK_FILE_EXT)
        lock = AutoURI(v6_txt).get_lock(no_lock=False)

        lock.acquire()
        try:
            assert u_lock.exists
            time.sleep(1)
            raise AutoURIFileLockTestException
        except AutoURIFileLockTestException:
            assert True
        else:
            assert False
        finally:
            lock.release()
            assert not u_lock.exists


def test_autouri_lock_with_context(local_v6_txt, gcs_v6_txt, s3_v6_txt):
    for v6_txt in (local_v6_txt, gcs_v6_txt, s3_v6_txt):
        u_lock = AutoURI(v6_txt + URIBase.LOCK_FILE_EXT)

        with AutoURI(v6_txt).get_lock(no_lock=False):
            assert u_lock.exists
            time.sleep(1)
        assert not u_lock.exists

        with AutoURI(v6_txt).get_lock(no_lock=True):
            assert not u_lock.exists
            time.sleep(1)
        assert not u_lock.exists


def test_autouri_lock_with_context_raise(local_v6_txt, gcs_v6_txt, s3_v6_txt):
    for v6_txt in (local_v6_txt, gcs_v6_txt, s3_v6_txt):
        u_lock = AutoURI(v6_txt + URIBase.LOCK_FILE_EXT)

        try:
            with AutoURI(v6_txt).get_lock(no_lock=False):
                assert u_lock.exists
                time.sleep(1)
                raise AutoURIFileLockTestException
        except AutoURIFileLockTestException:
            assert not u_lock.exists
        else:
            assert False


def test_autouri_lock_timeout(local_v6_txt):
    """Timeout = 3, 8 sec
    For local storage (AbsPath) only.
    Default poll_interval (10 sec) is too long for test remote files.
    """
    for v6_txt in (local_v6_txt,):
        u_lock = AutoURI(v6_txt + URIBase.LOCK_FILE_EXT)

        time_s = time.time()
        lock = AutoURI(v6_txt).get_lock(no_lock=False)
        lock.acquire()
        try:
            lock2 = AutoURI(v6_txt).get_lock(no_lock=False, timeout=3)
            lock2.acquire()
            try:
                pass
            finally:
                lock2.release()
        except Timeout:
            assert 2 < time.time() - time_s < 4
        else:
            assert False
        finally:
            lock.release()
        assert not u_lock.exists

        time_s = time.time()
        lock = AutoURI(v6_txt).get_lock(no_lock=False)
        lock.acquire()
        try:
            lock2 = AutoURI(v6_txt).get_lock(no_lock=False, timeout=8)
            lock2.acquire()
            try:
                pass
            finally:
                lock2.release()
        except Timeout:
            assert 7 < time.time() - time_s < 9
        else:
            assert False
        finally:
            lock.release()
        assert not u_lock.exists



"""S3 Bucket configuration:
    S3 Object versioning must be turned off
"""
import logging
from tempfile import NamedTemporaryFile
from typing import Optional, Tuple

import requests
from boto3 import client
from botocore.exceptions import ClientError
from filelock import BaseFileLock

from .autouri import AutoURI, URIBase
from .metadata import URIMetadata, get_seconds_from_epoch, parse_md5_str
from .ntp_now import now_utc

logger = logging.getLogger(__name__)


class S3URILock(BaseFileLock):
    """Locking without using S3 Object Lock.
    Without S3 object lock, boto3's put_object(), which is used for _write() and _cp() in this module,
    does not ensure consistency of multiple write operations at the same time.
    It overwrites for all write requests but the last object written.
    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.put_object

    To make this lock as stable as possible, this module uses .lock file with id(self) written on it.
    This module first checks if .lock does not exist, then tries to write .lock with id(self).
    It waits for a short time (self._lock_read_delay) and checks if written .lock has the same id(self).
    self._lock_read_delay is set as poll_interval/10.

    Class constants:
    - LOCK_FILE_EXPIRATION_SEC:
        Expiration of a lock file based on its modtime in seconds
        If expired then such lock file is ignored.
    """

    LOCK_FILE_EXPIRATION_SEC = 1800

    def __init__(self, lock_file, timeout=900, poll_interval=10.0, no_lock=False):
        super().__init__(lock_file, timeout=timeout)
        self._poll_interval = poll_interval
        self._lock_read_delay = self._poll_interval / 10.0

    def acquire(self, timeout=None, poll_interval=5.0):
        """To use self._poll_interval instead of poll_interval in args."""
        super().acquire(timeout=timeout, poll_interval=self._poll_interval)

    def _acquire(self):
        """Unlike GCSURI, this module does not use S3 Object locking.
        This will write id(self) on a .lock file.
        """
        u = S3URI(self._lock_file)
        str_id = str(id(self))
        try:
            if (
                not u.exists
                or now_utc().timestamp() > u.mtime + S3URILock.LOCK_FILE_EXPIRATION_SEC
            ):
                u.write(str_id, no_lock=True)
                self._lock_file_fd = id(self)

            elif u.read() == str_id:
                self._lock_file_fd = id(self)

        except ClientError as e:
            status = e.response["ResponseMetadata"]["HTTPStatusCode"]
            if status in (403, 404):
                raise
        return None

    def _release(self):
        u = S3URI(self._lock_file)
        try:
            u.rm(no_lock=True, silent=True)
            self._lock_file_fd = None
        except ClientError:
            pass
        return None


class S3URI(URIBase):
    """
    Class constants:
        LOC_PREFIX (inherited):
            Path prefix for localization. Inherited from URIBase class.
        DURATION_PRESIGNED_URL:
            Duration for presigned URLs in seconds.
        S3_COPY_OBJECT_FILE_SIZE_LIMIT:
            File size limit (in bytes) for S3 copy_object (for s3 to s3 transfer).

    Protected class constants:
        _CACHED_BOTO3_CLIENTS:
        _CACHED_PRESIGNED_URLS:
        _S3_PUBLIC_URL_FORMAT:
            End point for a bucket with public access + key path
    """

    DURATION_PRESIGNED_URL: int = 4233600
    S3_COPY_OBJECT_FILE_SIZE_LIMIT: int = 5 * 1024 * 1024 * 1024

    _CACHED_BOTO3_CLIENTS = {}
    _CACHED_PRESIGNED_URLS = {}
    _S3_PUBLIC_URL_FORMAT = "http://{bucket}.s3.amazonaws.com/{path}"

    _LOC_SUFFIX = ".s3"
    _SCHEMES = ("s3://",)

    def __init__(self, uri, thread_id=-1):
        super().__init__(uri, thread_id=thread_id)

    def _get_lock(self, timeout=None, poll_interval=None):
        if timeout is None:
            timeout = S3URI.LOCK_TIMEOUT
        if poll_interval is None:
            poll_interval = S3URI.LOCK_POLL_INTERVAL
        return S3URILock(
            self._uri + S3URI.LOCK_FILE_EXT,
            timeout=timeout,
            poll_interval=poll_interval,
        )

    def get_metadata(self, skip_md5=False, make_md5_file=False):
        exists, mt, sz, md5 = False, None, None, None

        cl = S3URI.get_boto3_client(self._thread_id)
        bucket, path = self.get_bucket_path()

        try:
            m = cl.head_object(Bucket=bucket, Key=path)["ResponseMetadata"][
                "HTTPHeaders"
            ]
            # make keys lower-case
            headers = {k.lower(): v for k, v in m.items()}
            exists = True

            if not skip_md5:
                if "content-md5" in headers:
                    md5 = parse_md5_str(headers["content-md5"])
                elif "etag" in headers:
                    md5 = parse_md5_str(headers["etag"])
                if md5 is None:
                    # make_md5_file is ignored for S3URI
                    md5 = self.md5_from_file

            if "content-length" in headers:
                sz = int(headers["content-length"])

            if "last-modified" in headers:
                mt = get_seconds_from_epoch(headers["last-modified"])

        except Exception:
            pass

        return URIMetadata(exists=exists, mtime=mt, size=sz, md5=md5)

    def read(self, byte=False):
        cl = S3URI.get_boto3_client(self._thread_id)
        bucket, path = self.get_bucket_path()

        obj = cl.get_object(Bucket=bucket, Key=path)
        if byte:
            return obj["Body"].read()
        return obj["Body"].read().decode()

    def find_all_files(self):
        cl = S3URI.get_boto3_client(self._thread_id)
        bucket, path = self.get_bucket_path()
        sep = S3URI.get_path_sep()
        if path:
            path = path.rstrip(sep) + sep

        result = []
        objs = cl.list_objects_v2(Bucket=bucket, Prefix=path).get("Contents")
        if objs:
            for obj in objs:
                scheme = S3URI.get_schemes()[0]
                uri = scheme + sep.join([bucket, obj["Key"]])
                result.append(uri)
        return result

    def _write(self, s):
        cl = S3URI.get_boto3_client(self._thread_id)
        bucket, path = self.get_bucket_path()

        if isinstance(s, str):
            b = s.encode("ascii")
        else:
            b = s
        cl.put_object(Bucket=bucket, Key=path, Body=b)
        return

    def _rm(self):
        cl = S3URI.get_boto3_client(self._thread_id)
        bucket, path = self.get_bucket_path()

        cl.delete_object(Bucket=bucket, Key=path)
        return

    def _cp(self, dest_uri):
        """Copy from S3URI to
        S3URI
        AbsPath
        """
        from .abspath import AbsPath

        dest_uri = AutoURI(dest_uri)
        cl = S3URI.get_boto3_client(self._thread_id)
        bucket, path = self.get_bucket_path()

        if isinstance(dest_uri, S3URI):
            dest_bucket, dest_path = dest_uri.get_bucket_path()

            if self.size >= S3URI.S3_COPY_OBJECT_FILE_SIZE_LIMIT:
                copy_function = cl.copy
            else:
                copy_function = cl.copy_object
            copy_function(
                CopySource={"Bucket": bucket, "Key": path},
                Bucket=dest_bucket,
                Key=dest_path,
            )
            return True

        elif isinstance(dest_uri, AbsPath):
            dest_uri.mkdir_dirname()
            with open(dest_uri._uri, "wb") as fp:
                cl.download_fileobj(Bucket=bucket, Key=path, Fileobj=fp)
            return True
        return False

    def _cp_from(self, src_uri):
        """Copy to S3URI from
        AbsPath
        HTTPURL
        """
        from .abspath import AbsPath
        from .httpurl import HTTPURL

        src_uri = AutoURI(src_uri)
        cl = S3URI.get_boto3_client(self._thread_id)
        bucket, path = self.get_bucket_path()

        if isinstance(src_uri, AbsPath):
            cl.upload_file(Filename=src_uri._uri, Bucket=bucket, Key=path)
            return True

        elif isinstance(src_uri, HTTPURL):
            r = requests.get(
                src_uri._uri,
                stream=True,
                allow_redirects=True,
                headers=requests.utils.default_headers(),
            )
            r.raise_for_status()
            with NamedTemporaryFile() as fp:
                for chunk in r.iter_content(HTTPURL.get_http_chunk_size()):
                    fp.write(chunk)
                fp.seek(0)
                cl.upload_fileobj(Fileobj=fp, Bucket=bucket, Key=path)
            return True
        return False

    def get_bucket_path(self) -> Tuple[str, str]:
        """Returns a tuple of URI's S3 bucket and path."""
        arr = self.uri_wo_scheme.split(S3URI.get_path_sep(), maxsplit=1)
        if len(arr) == 1:
            # root directory without path (key)
            bucket, path = arr[0], ""
        else:
            bucket, path = arr
        return bucket, path

    def get_presigned_url(self, duration=None, use_cached=False) -> str:
        """
        Args:
            duration: Duration in seconds. This is ignored if use_cached is on.
            use_cached: Use a cached URL.
        """
        cache = S3URI._CACHED_PRESIGNED_URLS
        if use_cached:
            if cache is not None and self._uri in cache:
                return cache[self._uri]
        cl = S3URI.get_boto3_client(self._thread_id)
        bucket, path = self.get_bucket_path()
        duration = duration if duration is not None else S3URI.DURATION_PRESIGNED_URL
        url = cl.generate_presigned_url(
            "get_object", Params={"Bucket": bucket, "Key": path}, ExpiresIn=duration
        )
        cache[self._uri] = url
        return url

    def get_public_url(self) -> str:
        bucket, path = self.get_bucket_path()
        return S3URI._S3_PUBLIC_URL_FORMAT.format(bucket=bucket, path=path)

    @staticmethod
    def get_boto3_client(thread_id=-1) -> client:
        if thread_id in S3URI._CACHED_BOTO3_CLIENTS:
            return S3URI._CACHED_BOTO3_CLIENTS[thread_id]
        else:
            cl = client("s3")
            S3URI._CACHED_BOTO3_CLIENTS[thread_id] = cl
            return cl

    @staticmethod
    def init_s3uri(
        loc_prefix: Optional[str] = None,
        duration_presigned_url: Optional[int] = None,
        s3_copy_object_file_size_limit: Optional[int] = None,
    ):
        if loc_prefix is not None:
            S3URI.LOC_PREFIX = loc_prefix
        if duration_presigned_url is not None:
            S3URI.DURATION_PRESIGNED_URL = duration_presigned_url
        if s3_copy_object_file_size_limit is not None:
            S3URI.S3_COPY_OBJECT_FILE_SIZE_LIMIT = s3_copy_object_file_size_limit



import os
import re
from pathlib import Path

import setuptools

META_PATH = Path("autouri", "__init__.py")
HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    with Path(HERE, *parts).open(encoding="utf-8") as f:
        return f.read()


META_FILE = read(META_PATH)


def find_meta(meta):
    """
    Extract __*meta*__ from META_FILE.
    """
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta), META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="autouri",
    version=find_meta("version"),
    python_requires=">=3.6",
    scripts=["bin/autouri"],
    author="Jin wook Lee",
    author_email="leepc12@gmail.com",
    description="Automatic localization for various URIs (s3://, gs://, http://, https:// and local path)",
    long_description="https://github.com/ENCODE-DCC/autouri",
    long_description_content_type="text/markdown",
    url="https://github.com/ENCODE-DCC/autouri",
    packages=setuptools.find_packages(exclude=["docs", "test"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
    ],
    install_requires=[
        "requests",
        "pyopenssl",
        "google-cloud-storage",
        "boto3",
        "awscli",
        "dateparser",
        "filelock>=3.4.0",
        "six>=1.13.0",
        "ntplib",
    ],
)



#!/usr/bin/env python3
import argparse
import logging
import os
import sys

from . import __version__ as version
from .autouri import AutoURI, URIBase
from .gcsuri import GCSURI
from .s3uri import S3URI

logger = logging.getLogger(__name__)


DEFAULT_RMDIR_NTH = 6


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument("-v", "--version", action="store_true", help="Show version")

    parent_all = argparse.ArgumentParser(add_help=False)
    group_log_level = parent_all.add_mutually_exclusive_group()
    group_log_level.add_argument(
        "-D", "--debug", action="store_true", help="Prints all logs >= DEBUG level"
    )

    parent_src = argparse.ArgumentParser(add_help=False)
    parent_src.add_argument("src", help="Source file URI")

    parent_lock = argparse.ArgumentParser(add_help=False)
    parent_lock.add_argument(
        "--no-lock", action="store_true", help="No file locking for write/rm action."
    )

    parent_target = argparse.ArgumentParser(add_help=False)
    parent_target.add_argument(
        "target",
        help="Target file/directory URI (e.g. gs://here/me.txt, s3://here/my-dir/) ."
        "Directory must have a trailing directory separator "
        "(e.g. /hello/. gs://where/am/i/).",
    )

    parent_cp = argparse.ArgumentParser(add_help=False)
    parent_cp.add_argument(
        "--use-gsutil-for-s3",
        action="store_true",
        help="Use gsutil for DIRECT TRANSFER between gs:// and s3://. "
        "gsutil must be installed and configured to have AWS credentials "
        'in ~/.boto file. Run "gsutil config" do generate it.',
    )
    parent_cp.add_argument(
        "--make-md5-file",
        action="store_true",
        help="Make .md5 file to store file's md5 hexadecimal string. "
        "This file can be used later to prevent repeated md5sum calculation. "
        "This is for local path only since cloud URIs already provide md5 hash "
        "info in HTTP headers.",
    )

    subparser = parser.add_subparsers(dest="action")

    subparser.add_parser(
        "metadata",
        help="AutoURI(src).get_metadata(): Get metadata of source.",
        parents=[parent_src, parent_all],
    )

    subparser.add_parser(
        "cp",
        help="AutoURI(src).cp(target): Copy source to target. "
        "target must be a full filename/directory. "
        "Target directory must have a trailing directory separator "
        "(e.g. /)",
        parents=[parent_src, parent_lock, parent_target, parent_cp, parent_all],
    )
    subparser.add_parser(
        "find",
        help="AutoURI(src).find_all_files(): Recursively list all files (not sub-directories) "
        "on source (directory).",
        parents=[parent_src, parent_all],
    )

    subparser.add_parser(
        "read",
        help="AutoURI(src).read(): Read from source.",
        parents=[parent_src, parent_all],
    )

    p_write = subparser.add_parser(
        "write",
        help="AutoURI(src).write(text): Write text on source.",
        parents=[parent_src, parent_lock, parent_all],
    )
    p_write.add_argument("text", help="Text to be written to source file.")

    subparser.add_parser(
        "rm",
        help="AutoURI(src).rm(): Delete source.",
        parents=[parent_src, parent_lock, parent_all],
    )

    p_rmdir = subparser.add_parser(
        "rmdir",
        help="AutoURI(src).rmdir(): Recursively delete all files on "
        "source directory.",
        parents=[parent_src, parent_lock, parent_all],
    )
    p_rmdir.add_argument("--delete", action="store_true", help="DELETE outputs.")
    p_rmdir.add_argument(
        "-t",
        "--num-threads",
        default=URIBase.DEFAULT_NUM_THREADS,
        type=int,
        help="Number of threads used for deleting "
        "multiple files on cloud buckets (gs://, s3://).",
    )

    p_loc = subparser.add_parser(
        "loc",
        help="AutoURI(src).localize_on(target): Localize source on target directory "
        "Target directory must end with directory separator",
        parents=[parent_src, parent_target, parent_cp, parent_lock, parent_all],
    )
    p_loc.add_argument(
        "--recursive",
        action="store_true",
        help="Recursively localize source into target directory.",
    )

    p_presign = subparser.add_parser(
        "presign",
        help="AutoURI(src).get_presigned_url(). For cloud-based URIs only.",
        parents=[parent_src, parent_all],
    )
    p_presign.add_argument(
        "--gcp-private-key-file",
        help="GCP private key file (JSON format) to presign gs:// URIs.",
    )
    p_presign.add_argument(
        "--duration", type=int, help="Duration of presigned URL in seconds."
    )

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()
    if args.version:
        print(version)
        parser.exit()

    if args.debug:
        log_level = "DEBUG"
    else:
        log_level = "INFO"
    logging.basicConfig(
        level=log_level, format="%(asctime)s|%(name)s|%(levelname)s| %(message)s"
    )
    # suppress filelock logging
    logging.getLogger("filelock").setLevel("CRITICAL")

    return args


def get_local_path_if_valid(s):
    abspath = os.path.abspath(os.path.expanduser(s))
    dirname = os.path.dirname(abspath)
    if os.path.isdir(abspath):
        return abspath + os.sep
    elif os.path.exists(abspath):
        return abspath
    elif os.path.exists(dirname):
        return abspath
    return s


def main():
    args = parse_args()

    src = get_local_path_if_valid(args.src)

    if args.action in ("cp", "loc"):
        if args.use_gsutil_for_s3:
            GCSURI.init_gcsuri(use_gsutil_for_s3=True)
        target = get_local_path_if_valid(args.target)

    if args.action == "metadata":
        m = AutoURI(src).get_metadata()
        print(m)

    elif args.action == "cp":
        u_src = AutoURI(src)
        _, flag = u_src.cp(
            target,
            make_md5_file=args.make_md5_file,
            return_flag=True,
            no_lock=args.no_lock,
        )

        if flag == 0:
            logger.info("Copying from file {s} to {t} done".format(s=src, t=target))
        elif flag:
            if flag == 1:
                reason = "skipped due to md5 hash match"
            elif flag == 2:
                reason = "skipped due to filename/size match and mtime test"
            else:
                raise NotImplementedError
            logger.info(
                "Copying from file {s} to {t} {reason}".format(
                    s=src, t=target, reason=reason
                )
            )
    elif args.action == "read":
        s = AutoURI(src).read()
        print(s)

    elif args.action == "find":
        for uri in AutoURI(src).find_all_files():
            print(uri)

    elif args.action == "write":
        AutoURI(src).write(args.text, no_lock=args.no_lock)
        logger.info("Text has been written to {s}".format(s=src))

    elif args.action == "rm":
        u = AutoURI(src)
        if not u.exists:
            raise ValueError("File does not exist. {s}".format(s=src))
        u.rm(no_lock=args.no_lock)
        logger.info("Deleted {s}".format(s=src))

    elif args.action == "rmdir":
        AutoURI(src).rmdir(
            dry_run=not args.delete, num_threads=args.num_threads, no_lock=args.no_lock
        )
        if not args.delete:
            logger.warning(
                "rmdir ran in a dry-run mode. "
                "Use --delete to DELETE ALL FILES on a directory."
            )

    elif args.action == "loc":
        _, localized = AutoURI(src).localize_on(
            target,
            recursive=args.recursive,
            make_md5_file=args.make_md5_file,
            return_flag=True,
            no_lock=args.no_lock,
        )
        if localized:
            logger.info("Localized {s} on {t}".format(s=src, t=target))
        else:
            logger.info("No need to localize {s} on {t}".format(s=src, t=target))

    elif args.action == "presign":
        u = AutoURI(src)

        if isinstance(u, GCSURI):
            if not args.gcp_private_key_file:
                raise ValueError(
                    "GCP private key file (--gcp-private-key-file) not found."
                )
            url = u.get_presigned_url(
                duration=args.duration, private_key_file=args.gcp_private_key_file
            )
            print(url)

        elif isinstance(u, S3URI):
            url = u.get_presigned_url(duration=args.duration)
            print(url)

        else:
            raise ValueError(
                "Presigning URL is available for cloud URIs (gs://, s3://) only."
            )


if __name__ == "__main__":
    main()
